module serialize4out

import StdEnv, StdMaybe, monad

/*
	Pieter Koopman, pieter@cs.ru.nl
	Advanced Programming, week 4, 2016
*/

// ---

:: State s a = S (s -> (Maybe a,s))

unS :: (State s a) -> s -> (Maybe a,s)
unS (S f) = f

instance Functor (State s) where
	fmap f (S g) = S \s.case g s of
		(Just a, s) = (Just (f a), s)
		(_,      s) = (Nothing   , s)
instance Applicative (State s) where
	pure a = S \s.(Just a,s)
	(<*>) (S f) (S x) = S \s.case f s of
		(Just f, s) = case x s of
			(Just x, s) = (Just (f x), s)
			(_     , s) = (Nothing, s)
		(_,      s) = (Nothing, s)
instance fail (State s) where
	fail = S \s.(Nothing, s)
instance Monad (State s) where
	bind (S a) f = S \s.case a s of
		(Just a, s) = unS (f a) s
		(_,      s) = (Nothing, s)
instance OrMonad (State s) where
	(<|>) (S f) (S g) = S \s.case f s of
		(Nothing, _) = g s
		other = other

// ---

:: Ser =
	{ ins :: [String]
	, out :: [String]
	}
ser :: Ser
ser = {ins=[], out = []}

toStrings :: Ser -> [String]
toStrings {ins,out} = ins ++ reverse ["\n":out]

:: Serialize a :== State Ser a

wrt :: a -> State Ser String | toString a
wrt a = S \s.(Just b,{s & out = [b:s.out]}) where b = toString a

rd :: State Ser String
rd = S r where
	r s=:{ins=[a:x]}     = (Just a,{s & ins = x})
	r s=:{ins=[],out=[]} = (Nothing, s)
	r    {ins=[],out=l}  = r {ins=reverse l,out = []}

match :: a -> Serialize a | toString a
match a = rd >>= \s.guard (toString a == s) >>| pure a <|> fail

pred :: (String->Bool) -> Serialize String
pred f = rd >>= \s.guard (f s) >>| pure s <|> fail

// ---

:: UNIT     = UNIT
:: EITHER a b = LEFT a | RIGHT b
:: PAIR   a b = PAIR a b
:: CONS   a   = CONS String a

:: Write a :== a -> Serialize String
:: Read a  :== Serialize a
 
class serialize a | isUNIT a where
  write :: a -> Serialize String
  read  :: Serialize a

class serialize1 t where
  write1 :: (Write a) (t a) -> Serialize String
  read1  :: (Read  a) -> Serialize (t a)

class serializeCONS a where
	writeCons :: (Write a) (CONS a) -> Serialize String
	readCons  :: String (Read a) -> Serialize (CONS a)

class serialize2 t where
  write2 :: (Write a) (Write b) (t a b) -> Serialize String
  read2  :: (Read  a) (Read  b) -> Serialize (t a b)

class isUNIT a :: a -> Bool
instance isUNIT UNIT where isUNIT _ = True
instance isUNIT a    where isUNIT _ = False

instance serialize Bool where
  write b = wrt b
  read = match True <|> match False

instance serialize Int where
	write i = wrt i
	read = rd >>= \s.pure (toInt s) >>= \i.guard (s == toString i) >>| pure i

instance serialize UNIT where
	write _ = pure ""
	read = pure UNIT

instance serializeCONS UNIT where
	writeCons wa (CONS name a) = wrt name
	readCons name ra = CONS <$> match name <*> pure UNIT
 
instance serializeCONS a where
	writeCons wa (CONS name a) = wrt "(" >>| wrt name >>| wa a >>| wrt ")"
	readCons name ra = match "(" >>| CONS <$> match name <*> ra >>= \c. match ")" >>| pure c
 
instance serialize2 EITHER where
  write2 wa wb (LEFT  a) = wa a
  write2 wa wb (RIGHT b) = wb b
  read2 ra rb = LEFT <$> ra <|> RIGHT <$> rb

instance serialize2 PAIR where
  write2 wa wb (PAIR a b) = wa a >>| wrt " " >>| wb b
  read2 ra rb = PAIR <$> ra <*> (match " " >>| rb)

// ---

:: ListG a :== EITHER (CONS UNIT) (CONS (PAIR a [a]))

fromList :: [a] -> ListG a
fromList []  = LEFT  (CONS NilString  UNIT)
fromList [a:x] = RIGHT (CONS ConsString (PAIR a x))

toList :: (ListG a) -> [a]
toList (LEFT  (CONS NilString  UNIT)) = []
toList (RIGHT (CONS ConsString (PAIR a x))) = [a:x]

NilString :== "Nil"
ConsString :== "Cons"

instance serialize [a] | serialize a where
 write a = write1 write a
 read    = read1  read

instance serialize1 [] where
	write1 writea l = write2 (writeCons write) (writeCons (write2 writea (write1 writea))) (fromList l)
	read1  reada
	 = toList <$> read2 (readCons NilString read) (readCons ConsString (read2 reada (read1 reada)))

// ---

:: Bin a = Leaf | Bin (Bin a) a (Bin a)

:: BinG a :== EITHER (CONS UNIT) (CONS (PAIR (Bin a) (PAIR a (Bin a))))

fromBin :: (Bin a) -> BinG a
fromBin Leaf = LEFT (CONS LeafString UNIT)
fromBin (Bin l a r) = RIGHT (CONS BinString (PAIR l (PAIR a r)))

toBin :: (BinG a) -> Bin a
toBin (LEFT (CONS _ UNIT)) = Leaf
toBin (RIGHT (CONS _ (PAIR l (PAIR a r)))) = Bin l a r

LeafString :== "Leaf"
BinString :== "Bin"

instance == (Bin a) | == a where
  (==) Leaf Leaf = True
  (==) (Bin l a r) (Bin k b s) = l == k && a == b && r == s
  (==) _ _ = False

instance serialize (Bin a) | serialize a where
	write b = write1 write b
	read = read1 read

instance serialize1 Bin where
	write1 writea b = write2 (writeCons write) (writeCons (write2 (write1 writea) (write2 writea (write1 writea)))) (fromBin b)
	read1  reada    = toBin <$> read2 (readCons LeafString read) (readCons BinString (read2 (read1 reada) (read2 reada (read1 reada))))

// ---

:: Coin = Head | Tail
:: CoinG :== EITHER (CONS UNIT) (CONS UNIT)

fromCoin :: Coin -> CoinG
fromCoin Head = LEFT (CONS "Head" UNIT)
fromCoin Tail = RIGHT (CONS "Tail" UNIT)

toCoin :: CoinG -> Coin
toCoin (LEFT (CONS _ UNIT)) = Head
toCoin (RIGHT (CONS _ UNIT)) = Tail

instance == Coin where
  (==) Head Head = True
  (==) Tail Tail = True
  (==) _    _    = False

instance serialize Coin where
	write c = write2 (writeCons write) (writeCons write) (fromCoin c)
	read    = toCoin <$> read2 (readCons "Head" read) (readCons "Tail" read)

// ---

Start = 
  [test True
  ,test False
  ,test 0
  ,test 123
  ,test -36
  ,test Head
  ,test Tail
  ,test [42]
  ,test [0..4]
  ,test [[True],[]]
  ,test (Bin Leaf True Leaf)
  ,test [Bin (Bin Leaf [1] Leaf) [2] (Bin Leaf [3] (Bin Leaf [4,5] Leaf))]
  ,test [Bin (Bin Leaf [1] Leaf) [2] (Bin Leaf [3] (Bin (Bin Leaf [4,5] Leaf) [6,7] (Bin Leaf [8,9] Leaf)))]
  ,test Head
  ,test Tail
  ,["End of the tests.\n"]
  ]

test :: a -> [String] | serialize, == a
test a = toStrings (snd ((unS t) ser)) where
 t
 	=   write a
	>>| read
	>>= \b. guard (a == b)
	>>| wrt "Oke "
	<|> wrt "Failure "

