module ideabox16

import iTasks, StdArray

Start world = startEngine (nameTask mainTask) world

:: Idea =
  { idea_name :: String
  , idea_description :: Maybe Note
  }

:: NamedIdea =
  { number :: Int
  , idea   :: Idea
  , user   :: String
  }

:: DisplayIdea =
  { num :: Int
  , idea_title :: String
  , owner :: String
  }

derive class iTask Idea, NamedIdea, DisplayIdea

ideas :: Shared [NamedIdea]
ideas = sharedStore "Ideas" []

nameTask :: (String -> Task a) -> Task a | iTask a
nameTask task
 =    updateInformation "enter your name" [] ""
  >>* [OnAction ActionOk (ifValue (\name.size name > 2) task)]

mainTask :: String -> Task [NamedIdea]
mainTask user | user == "admin"
  = updateSharedInformation "admin view" [] ideas
  = userTask user

userTask :: String -> Task [NamedIdea]
userTask user
  =   get ideas >>= \list.
      (enterChoiceWithShared "Current idea list"
        [ChooseWith (ChooseFromGrid (\{number,idea,user}.{num=number,idea_title=idea.idea_name,owner=user}))]
//        viewSharedInformation "plans" []
        ideas >&^ viewSharedInformation "selected idea" []
        )
  ||- enterInformation ("enter a new idea " +++ user) []
      >>* [ OnAction ActionOk (ifValue (valid list) addIdea)
          , OnAction (Action "Clear" [ActionKey (unmodified KEY_BACKSPACE)]) (always (userTask user))
          , OnAction ActionQuit (always (get ideas))
          , OnAction ActionDelete (always check)
          ]
where
  valid list idea = size idea.idea_name > 2 && not (isMember idea.idea_name (map (\i=i.idea.idea_name) list))
  addIdea idea 
   =  upd (\list. list ++ [{ idea = idea, user = user, number = length list}]) ideas
    >>| userTask user
  check
    = viewInformation "are you sure to delete all ideas?" [] ""
    >>* [ OnAction ActionYes (always (upd (\list.[]) ideas))
        , OnAction ActionNo  (always (get ideas))
        ]
    >>| userTask user

/*
userTask :: String -> Task [NamedIdea]
userTask user
  = forever (
      get ideas >>= \list.
      (enterChoiceWithShared "Current idea list"
        [ChooseWith (ChooseFromGrid (\{number,idea,user}.{num=number,idea_title=idea.idea_name,owner=user}))]
//        viewSharedInformation "plans" []
        ideas >&^ viewSharedInformation "selected idea" []
        )
  ||- enterInformation ("enter a new idea " +++ user) []
      >>* [ OnAction ActionOk (ifValue (valid list) addIdea)
          , OnAction (Action "Clear" [ActionKey (unmodified KEY_BACKSPACE)]) (always (get ideas))
          , OnAction ActionQuit (always (throw 0))
          , OnAction ActionDelete (always check)
          ]
          )
where
  valid list idea = size idea.idea_name > 2 && not (isMember idea.idea_name (map (\i=i.idea.idea_name) list))
  addIdea idea 
   =  upd (\list. list ++ [{ idea = idea, user = user, number = length list}]) ideas
  check
    = viewInformation "are you sure to delete all ideas?" [] ""
    >>* [ OnAction ActionYes (always (upd (\list.[]) ideas))
        , OnAction ActionNo  (always (get ideas))
        ]
*/
