module rail

import iTasks
import iTasks.API.Extensions.SVG.SVGlet

Start :: *World -> *World
Start world = startEngine [publish "/" (WebApp []) (\_ -> roleTask)] world

network :: Shared Network
network = sharedStore "network" {elements = [], signals = [], trains = []}

signals :: [Element] -> [SignalState]
signals [Section s:x]
 = if s.sLeftSignal  [(Display (leftLabel  s.sLabel),False)] [] ++
   if s.sRightSignal [(Display (rightLabel s.sLabel),False)] [] ++
   signals x
signals [Point p:x]
 = [(Display p.pLabel,False):signals x]
signals l = []

leftLabel l = l + "_Left"
rightLabel l = l + "_Right"

readSignal :: String [SignalState] -> Bool
readSignal s [(Display l,b):x] | s == l
  = b
  = readSignal s x
readSignal s [] = False

writeSignal :: String [SignalState] -> [SignalState]
writeSignal s [(Display l,b):x] | s == l
  = [(Display l, not b):x]
  = [(Display l, b): writeSignal s x]
writeSignal s [] = []

roleTask :: Task Network
roleTask =
	get network >>= \net.
  viewInformation "select role" [] ""
  >>*([OnAction (Action "Design" []) (always designTask)
      ,OnAction (Action "Controller" []) (always controlTask)
//      ,OnAction (Action "All" []) (always task)
      ] ++
      [OnAction (Action ("Driver " + t.driver) []) (always (driverTask t.driver))
      \\ t <- net.trains
      ]
     )

designTask :: Task Network
designTask =
  (       (elementsTask <<@ ArrangeHorizontal)
  -||-  (imageTask False <<@ ArrangeHorizontal)
  ) <<@ ArrangeHorizontal

controlTask :: Task Network
controlTask =
  (         (signalTask <<@ ArrangeHorizontal)
  -||-	(makeTrainTask <<@ ArrangeHorizontal)
  -||-  (imageTask True <<@ ArrangeHorizontal)
  ) <<@ ArrangeHorizontal

task :: Task Network
task =
  (    (elementsTask <<@ ArrangeHorizontal)
    -||- (signalTask   <<@ ArrangeHorizontal)
  -||-	(makeTrainTask <<@ ArrangeHorizontal)
  -||- (imageTask True <<@ ArrangeHorizontal)
  ) <<@ ArrangeHorizontal

elementsTask =
  updateSharedInformation
    (Title "Elements")
    [UpdateWithShared (\n.n.elements) (\n e.setStates {n&elements = e, signals = signals e}) (\v1 v2.v1)]
    network

signalTask =
  updateSharedInformation
    (Title "Signals")
    [UpdateWithShared (\n.n.signals) (\n s.setStates {n&signals = s}) (\n1 n2.n1)]
    network

imageTask update =
  updateSharedInformation
    (Title "Layout")
    [imageUpdate
      id
      (\s v tags -> image s.signals update s)
      (\s v -> setStates v) //{s & signals = v.signals})
      (\s v -> s)
      (\_ s -> Nothing)
      (\o n.n)
    ]
    network

makeTrainTask :: Task Network
makeTrainTask =
	updateSharedInformation
		(Title "Trains")
		[UpdateWithShared (\n.n.trains) (\n t.setStates {n & trains = t}) (\x y.x)]
		network

driverTask :: String -> Task Network
driverTask name =
	((forever
	(	(viewSharedInformation
			(Title ("Train driven by " + name))
			[ViewWith (\n.findTrain name n.trains)]
			network)
		>>* [OnAction (Action "Go" []) (always (upd (setState name o stepTrain name) network))]
	) <<@ ArrangeHorizontal)
  -||-  (imageTask False <<@ ArrangeHorizontal))
  <<@ ArrangeHorizontal

setStates :: Network -> Network
setStates n = {n & trains = [{t & trainState = newState t n}\\ t <- n.trains]}

setState :: String Network -> Network
setState name net
	= case findTrain name net.trains of
		Just train = {net & trains = updateTrain {train & trainState = newState train net} net.trains}	
		_ = net

newState :: Train Network -> TrainState
newState train net
	| isJust pointBelow
		| readSignal (fromJust pointBelow).pLabel net.signals
			= Go
			= Derailed
	| isJust pointAbove
		| readSignal (fromJust pointAbove).pLabel net.signals
			= Go
			= Derailed
	| isNothing mbelement
		= Derailed
	| train.current === train.target
		= Arrived
		= case fromJust mbelement of
			(Section s)
				| train.toRight
					| s.sRightSignal
						| readSignal (rightLabel s.sLabel) net.signals
							= Go
							= Wait
						= Go
					| s.sLeftSignal
						| readSignal (leftLabel s.sLabel) net.signals
							= Go
							= Wait
						= Go
			(Point p) //| readSignal p.pLabel net.signals
				//= Derailed
				= Go
where
	pos = train.current
	mbelement = elementAt pos net.elements
	pointBelow = case elementAt {pos & y = pos.y + 1} net.elements of
								Just (Point p) | p.pOrientation === NE || p.pOrientation === NW
									= Just p
								_ = Nothing
	pointAbove = case elementAt {pos & y = pos.y - 1} net.elements of
								Just (Point p) | p.pOrientation === SE || p.pOrientation === SW
									= Just p
								_ = Nothing

stepTrain :: String Network -> Network
stepTrain name net
 = case findTrain name net.trains of
 		Just train | train.trainState === Go
 			= {net & trains = updateTrain (step train) net.trains}
 			= net
 		_ = net
where
	step train | train.toRight
		# pos = train.current
		# pos = {pos & x = pos.x + 1}
		# pos = case elementAt pos net.elements of
							Just (Point p) | readSignal p.pLabel net.signals
								= case p.pOrientation of
									SE = {pos & y = pos.y + 1}
									NE = {pos & y = pos.y - 1}
									_  = pos
							_ = pos
		# posY1 = {pos & y = pos.y + 1}
		# pos = case elementAt posY1 net.elements of
							Just (Point p) | readSignal p.pLabel net.signals
								= case p.pOrientation of
									NW = posY1
									_  = pos
							_ = pos
		# posY1 = {pos & y = pos.y - 1}
		# pos = case elementAt posY1 net.elements of
							Just (Point p) | readSignal p.pLabel net.signals
								= case p.pOrientation of
									SW = posY1
									_  = pos
							_ = pos
		= {train & current = pos}
		// to the left
		# pos = train.current
		# pos = {pos & x = pos.x - 1}
		# pos = case elementAt pos net.elements of
							Just (Point p) | readSignal p.pLabel net.signals
								= case p.pOrientation of
									SW = {pos & y = pos.y + 1}
									NW = {pos & y = pos.y - 1}
									_  = pos
							_ = pos
		# posY1 = {pos & y = pos.y + 1}
		# pos = case elementAt posY1 net.elements of
							Just (Point p) | readSignal p.pLabel net.signals
								= case p.pOrientation of
									NE = posY1
									_  = pos
							_ = pos
		# posY1 = {pos & y = pos.y - 1}
		# pos = case elementAt posY1 net.elements of
							Just (Point p) | readSignal p.pLabel net.signals
								= case p.pOrientation of
									SE = posY1
									_  = pos
							_ = pos
		= {train & current = pos}

findTrain :: String [Train] -> Maybe Train
findTrain name [] = Nothing
findTrain name [a:x] | a.driver == name
	= Just a
	= findTrain name x 

updateTrain :: Train [Train] -> [Train]
updateTrain t l = [if (t.driver == e.driver) t e \\ e <- l]

:: ELabel :== String
:: Position = {x :: Int, y :: Int}
:: Element = Section Section | Point Point
:: Orientation = NE | SE | SW | NW
:: Section =
  { sLabel       :: ELabel
  , sPosition    :: Position
  , sLeftSignal  :: Bool
  , sRightSignal :: Bool
  }
:: SignalState :== (Display String, Bool)
:: Point =
  { pLabel       :: ELabel
  , pPosition    :: Position
  , pOrientation :: Orientation
  }
:: Train =
	{ driver	:: String
	, current	:: Position
	, target	:: Position
	, toRight	:: Bool
	, trainState :: TrainState
	}
:: TrainState = Wait | Go | Arrived | Derailed
:: Network =
  { elements :: [Element]
  , signals  :: [SignalState]
  , trains   :: [Train]
  }

element :: ELabel [Element] -> Maybe Element
element e [] = Nothing
element e [a:x] | elementLabel a == e
  = Just a
  = element e x

elementLabel :: Element -> ELabel
elementLabel (Section s) = s.sLabel
elementLabel (Point   p) = p.pLabel

elementAt :: Position [Element] -> Maybe Element
elementAt pos [] = Nothing
elementAt pos [Section s:x] | s.sPosition === pos
	= Just (Section s)
	= elementAt pos x
elementAt pos [Point p:x] | p.pPosition === pos
	= Just (Point p)
	= elementAt pos x	

derive class iTask Position, Element, Orientation, Section, Point, Network, Train, TrainState

// visualisation
font = normalFontDef "Arial" fontSize
fontSize = 12.0

boxSize = px 80.0
boxHeight = px 50.0
signalSize = boxSize /. 8
lineWidth = px 3.0
boxWidth = px 1.5

:: StateTrans :== Int -> Int

class image a :: [SignalState] Bool a -> Image Network
instance image Element where
  image sig update (Section s) = image sig update s
  image sig update (Point p)   = image sig update p
instance image Section where
  image sig update s
   =  margin (boxHeight *. s.sPosition.y, px0, px0, boxSize *. s.sPosition.x)
      (overlay [(AtMiddleX,AtBottom)] []
        ([above (repeat AtMiddleX) []
          [xline Nothing boxSize <@< {strokewidth = lineWidth}
          ,label s.sLabel
          ] Nothing
        ] ++
        if s.sLeftSignal
          [let l = leftLabel s.sLabel in
           margin (signalSize /. 2, px0, px0, signalSize /. 2) (drawSignal (readSignal l sig) update l)]
          [] ++
        if s.sRightSignal
          [let l = rightLabel s.sLabel in
           margin (signalSize /. 2, px0, px0, boxSize - signalSize *. 1.5) (drawSignal (readSignal l sig) update l)]
          []
        )
       (Just (box 1 1)))
instance image Point where
  image sig update p
   = margin (boxHeight *. (p.pPosition.y - 1 + case p.pOrientation of SE = 1; SW = 1; _ = 0), px0, px0, boxSize *. p.pPosition.x)
      (overlay [(AtMiddleX,AtBottom)] []
        [above (repeat AtMiddleX) []
        (case p.pOrientation of
          NE =
            [line Nothing Slash boxSize boxHeight
              <@< {strokewidth = lineWidth}
              <@< {dash = if B [6,3] []}
            ,xline Nothing boxSize
              <@< {strokewidth = lineWidth}
              <@< {dash = if b [6,3] []}
            ,label p.pLabel
            ]
          NW = 
            [line Nothing Backslash boxSize boxHeight
              <@< {strokewidth = lineWidth}
              <@< {dash = if B [6,3] []}
            ,xline Nothing boxSize
              <@< {strokewidth = lineWidth}
              <@< {dash = if b [6,3] []}
            ,label p.pLabel
            ]
          SE =
            [xline Nothing boxSize
              <@< {strokewidth = lineWidth}
              <@< {dash = if b [6,3] []}
            ,line Nothing Backslash boxSize boxHeight
              <@< {strokewidth = lineWidth}
              <@< {dash = if B [6,3] []}
            ,label p.pLabel
            ]
          SW =
            [xline Nothing boxSize
              <@< {strokewidth = lineWidth}
              <@< {dash = if b [6,3] []}
            ,line Nothing Slash boxSize boxHeight
              <@< {strokewidth = lineWidth}
              <@< {dash = if B [6,3] []}
            ,label p.pLabel
            ]
        ) Nothing
        ]
        (Just (box 1 2 <@< {onclick = \i n.if update (updateState {n & signals = writeSignal p.pLabel n.signals}) n, local = False})))
  where
    b = readSignal p.pLabel sig
    B = not b
    
updateState :: Network -> Network
updateState n = {n & trains = [{t & trainState = newState t n} \\ t <- n.trains]} 

instance image Train where
	image sig update t = train t.current t.toRight <@< {onclick = \i n.stepTrain t.driver n, local = False}
instance image Network where
  image sig update n = margin (px 20.0) (collage []  (map (image n.signals update) n.elements ++ map (image n.signals update) n.trains) Nothing)

label :: ELabel -> Image m
label s = margin lineWidth (text font s)

box x y = 
  rect (boxSize *. x) (boxHeight *. y)
    <@< {fill = toSVGColor "aqua"}
    <@< {strokewidth = boxWidth}
    <@< {stroke = toSVGColor "lightgrey"}

drawSignal b update l
    = circle signalSize 
    <@< {fill = toSVGColor (if b "lightgreen" "red")} 
    <@< {strokewidth = px 2.0}
    <@< {onclick = \i n.if update (updateState {n & signals = writeSignal l n.signals}) n,local = False}
/*
bitmap x y w h img
 = raw
    (px (toReal w))
    (px (toReal h))
    ("<foreignObject x=\"" + toString x + "\" y=\"" + toString y + 
      "\" width=\"" + toString w + "\" height=\"" + toString h +
      "\"> <img src=\"" + img + "\" /></foreignObject>")
*/
train :: Position Bool -> Image a
train p r =
  margin (boxHeight *. p.y, px0, px0, boxSize *. p.x)
  (overlay [(AtMiddleX,AtTop)] []
    [if r flipx id 
    		(above (repeat AtMiddleX) [] [house,wheels] Nothing)
    ]
    (Just (empty boxSize boxHeight)))

px0 = px 0.0

house :: Image a
house
 = polygon Nothing [(zero,x),(zero,x2),(x2,x2),(x2,zero),(x,zero),(x,x),(x3,x),(x3,x4),(x4,x4),(x4,x)]
   <@< {fill = toSVGColor "grey"}
where
  x = px 12.0
  x2 = px 24.0
  x3 = px 8.0
  x4 = px 4.0

wheels :: Image a
wheels = beside [] [] [wheel, wheelSpace, wheel, wheelSpace, wheel] Nothing
//wheels = beside [] [(wheelSpaceSize *. n,zero) \\ n <- [0..]] (repeatn wheelCount wheel) (Just (empty wheelSize (wheelSize *. wheelCount + wheelSpaceSize *. (wheelCount - 1))))

wheel :: Image a
wheel = circle wheelSize

wheelSpace :: Image a
wheelSpace = empty wheelSpaceSize wheelSpaceSize

wheelSize = px 6.0
wheelSpaceSize = px 2.0
wheelCount = 3

unravel :: .[.a] -> (.[.a],.[.a])
unravel [a,b:abs]   = ([a:as],[b:bs]) where (as,bs) = unravel abs

instance + String where (+) s t = s +++ t

seq :: [a->a] a -> a
seq [f:fs] a = seq fs (f a)
seq [] a = a


